#' @title Generates synthetic data according to the model implemented in BASiCS
#'
#' @description \code{BASiCS_Sim} creates a simulated dataset from the model 
#' implemented in BASiCS.
#'
#' @param Mu Gene-specific mean expression parameters \eqn{\mu_i} for all 
#' biological genes (vector of length \code{q.bio}, 
#' all elements must be positive numbers)
#' @param Mu_spikes \eqn{\mu_i} for all technical genes defined as true 
#' input molecules (vector of length \code{q-q.bio}, 
#' all elements must be positive numbers)
#' @param Delta Gene-specific biological over-dispersion parameters 
#' \eqn{\delta_i}, biological genes only
#' (vector of length \code{q.bio}, all elements must be positive numbers)
#' @param Phi Cell-specific mRNA content normalising parameters \eqn{\phi_j}
#' (vector of length \code{n}, all elements must be positive numbers and 
#' the sum of its elements must be equal to \code{n})
#' @param S Cell-specific technical normalising parameters \eqn{s_j}
#' (vector of length \code{n}, all elements must be positive numbers)
#' @param Theta Technical variability parameter \eqn{\theta} (must be positive)
#'
#' @return An object of class \code{\linkS4class{SingleCellExperiment}}, 
#' including synthetic data generated by the model implemented in BASiCS.
#'
#' @examples
#'
#' # Simulated parameter values for 10 genes
#' # (7 biogical and 3 spike-in) measured in 5 cells
#' Mu <- c(8.36, 10.65, 4.88, 6.29, 21.72, 12.93, 30.19)
#' Mu_spike <-  c(1010.72, 7.90, 31.59)
#' Delta <- c(1.29, 0.88, 1.51, 1.49, 0.54, 0.40, 0.85)
#' Phi <- c(1.00, 1.06, 1.09, 1.05, 0.80)
#' S <- c(0.38, 0.40, 0.38, 0.39, 0.34)
#' Theta <- 0.39
#'
#' Data <- BASiCS_Sim(Mu, Mu_spike, Delta, Phi, S, Theta)
#' head(counts(Data))
#' dim(counts(Data))
#' metadata(Data)$SpikeInput
#' SingleCellExperiment::isSpike(Data)
#'
#' @author Catalina A. Vallejos \email{cnvallej@@uc.cl}, Nils Eling
#'
#' @references 
#' 
#' Vallejos, Marioni and Richardson (2015). PLoS Computational Biology. 
#' 
#' @export
BASiCS_Sim <- function(Mu, Mu_spikes, Delta, Phi, S, Theta) 
{
  # Number of cells
  n <- length(Phi)
  # Total number of genes, including biological and technical ones
  q <- length(Mu) + length(Mu_spikes)
  # Number of biological genes
  q.bio <- length(Delta)
    
  # Arguments checking
  if (!(is.vector(Mu) & is.numeric(Mu) & all(Mu > 0))) 
    stop("Invalid argument value for 'Mu'.")
  if (!(is.vector(Mu_spikes) & is.numeric(Mu_spikes) & all(Mu_spikes > 0))) 
    stop("Invalid argument value for 'Mu_spikes'.")
  if (!(is.vector(Delta) & is.numeric(Delta) & all(Delta >= 0))) 
    stop("Invalid argument value for 'Delta'.")
  if (!(is.vector(Phi) & is.numeric(Phi) & all(Phi > 0) & 
        isTRUE(all.equal(sum(Phi), n, tolerance = 0.01)))) 
    stop("Invalid argument value for 'Phi'.")
  if (!(is.vector(S) & is.numeric(S) & all(S > 0) & length(S) == n)) 
    stop("Invalid argument value for 'S'.")
  if (!(is.numeric(Theta) & length(Theta) == 1 & Theta >= 0)) 
    stop("Invalid argument value for 'Theta'.")
    
  if (!all(c(n, q, q.bio, q - q.bio) > 0)) 
    stop("Arguments' dimensions are not compatible")
    
  # Merge biological and technical genes
  Mu <- c(Mu, Mu_spikes)
    
  # Matrix where simulated counts will be stored
  Counts.sim <- matrix(0, ncol = n, nrow = q)
  # Matrix where gene-cell specific simulated random effects will be stored
  rho <- matrix(1, ncol = n, nrow = q.bio)
  # Simulated cell-specific random effects
  if (Theta > 0) 
  {
    Nu <- rgamma(n, shape = 1/Theta, rate = 1/(S * Theta))
  } 
  else 
  {
    Nu <- S
  }
  # Simulated counts data
  for (i in seq_len(q)) 
  {
    # Biological genes
    if (i <= q.bio) 
    {
      if (Delta[i] > 0) 
      {
        rho[i, ] <- rgamma(n, shape = 1/Delta[i], rate = 1/Delta[i])
      }
      Counts.sim[i, ] <- rpois(n, lambda = Phi * Nu * rho[i, ] * Mu[i])
    } 
    else 
    {
      # Technical genes
        Counts.sim[i, ] <- rpois(n, lambda = Nu * Mu[i])
    }
  }
  
  rownames(Counts.sim) <- paste0("Gene", seq_len(q))
  SpikeInfo <- data.frame(paste0("Gene", seq(q.bio+1,q)), Mu[seq(q.bio+1,q)])
  Data <- newBASiCS_Data(Counts = Counts.sim, Tech = ifelse(seq_len(q) > q.bio, 
                                                            TRUE, FALSE), 
                         SpikeInfo = SpikeInfo)
    
  return(Data)
}
