#' @name BASiCS_Chain
#' @aliases BASiCS_Chain-class
#' 
#' @title The BASiCS_Chain class
#' 
#' @description Container of an MCMC sample of the BASiCS' model parameters 
#' as generated by the function \code{\link[BASiCS]{BASiCS_MCMC}}. 
#' @slot parameters List of matrices containing MCMC chains for each model 
#' parameter. Depending on the mode in which BASiCS was run, the following 
#' parameters can appear in the list:
#' \describe{
#' \item{\code{mu}}{MCMC chain for gene-specific mean expression parameters 
#' \eqn{\mu_i}, biological genes only 
#' (matrix with \code{q.bio} columns, all elements must be positive numbers)}
#' \item{delta}{MCMC chain for gene-specific biological over-dispersion 
#' parameters \eqn{\delta_i}, biological genes only 
#' (matrix with \code{q.bio} columns, all elements must be positive numbers)}
#' \item{phi}{MCMC chain for cell-specific mRNA content normalisation parameters 
#' \eqn{\phi_j} (matrix with \code{n} columns, all elements must be positive 
#' numbers and the sum of its elements must be equal to \code{n})}
#' This parameter is only used when spike-in genes are available.
#' \item{s}{MCMC chain for cell-specific technical normalisation parameters 
#' \eqn{s_j} (matrix with \code{n} columns, 
#' all elements must be positive numbers)}
#' \item{nu}{MCMC chain for cell-specific random effects \eqn{\nu_j}
#' (matrix with \code{n} columns, all elements must be positive numbers)}
#' \item{theta}{MCMC chain for technical over-dispersion parameter(s) 
#' \eqn{\theta} (matrix, all elements must be positive, 
#' each colum represents 1 batch)}
#' \item{\code{beta}}{Only relevant for regression BASiCS model (Eling et al, 
#' 2017). MCMC chain for regression coefficients (matrix with \code{k} columns, 
#' where \code{k} represent the number of chosen basis functions + 2) }
#' \item{\code{sigma2}}{Only relevant for regression BASiCS model (Eling et al, 
#' 2017). MCMC chain for the residual variance (matrix with one column, sigma2 
#' represents a global parameter)}
#' \item{\code{epsilon}}{Only relevant for regression BASiCS model (Eling et al, 
#' 2017). MCMC chain for the gene-specific residual over-dispersion parameter 
#' (matrix with \code{q} columns)}
#' \item{\code{RefFreq}}{Only relevant for no-spikes BASiCS model (Eling et al, 
#' 2017). For each biological gene, this vector displays the proportion of 
#' times for which each gene was used as a reference (within the MCMC 
#' algorithm), when using the stochastic reference choice described in 
#' (Eling et al, 2017). This information has been kept as it is useful for the
#' developers of this library. However, we do not expect users to need it. }
#' }
#' 
#' @examples
#' 
#' # A BASiCS_Chain object created by the BASiCS_MCMC function.
#' Data <- makeExampleBASiCS_Data()
#' 
#' # To run the model without regression
#' Chain <- BASiCS_MCMC(Data, N = 100, Thin = 2, Burn = 2, Regression = FALSE)
#' 
#' # To run the model using the regression model
#' ChainReg <- BASiCS_MCMC(Data, N = 100, Thin = 2, Burn = 2, Regression = TRUE)
#' 
#' @author Catalina A. Vallejos \email{cnvallej@@uc.cl}
#' @author Nils Eling \email{eling@@ebi.ac.uk}
#' 
#' @export
setClass("BASiCS_Chain", 
         representation = representation(
           parameters = "list"), 
         contains = "Versioned",
         validity = function(object) 
         {
            errors <- character()

            if(!("parameters" %in% methods::slotNames(object)))
            {
              errors <- "'object' was generated by an old version of BASiCS"
              return(errors)
            }
            else
            {
              ValidNames <- c("mu",
                              "delta",
                              "phi",
                              "s",
                              "nu",
                              "theta",
                              "beta",
                              "sigma2",
                              "epsilon",
                              "RefFreq",
                              "designMatrix",
                              "ls.mu",
                              "ls.delta",
                              "ls.phi",
                              "ls.nu",
                              "ls.theta")
              ReqNames <- c("mu",
                            "delta",
                            "s",
                            "nu",
                            "theta")
              
              # Check whether all elements of `parameters` are valid
              if(sum(!(names(object@parameters) %in% ValidNames) > 0) > 0)
                errors <- c(errors,
                  paste("Invalid elements",
                    paste(setdiff(names(object@parameters), ValidNames), collapse = ", "),
                    "in `parameters` slot"))
              
              # Check whether all minimum `parameters` are present
              if(sum(names(object@parameters) %in% ReqNames) != 5)
                errors <- c(errors, "One or more parameters are missing")     
              
              # Check for infinite values and NAs
              if(sum(lapply(
                object@parameters[names(object@parameters) != "epsilon"], 
                function(x) sum(!is.finite(x))) > 0) > 0)
                errors <- c(errors, "Some parameters contain NA/Inf values")
              
              N <- nrow(object@parameters$mu)
              q <- ncol(object@parameters$mu)
              n <- ncol(object@parameters$s)
              
              # Check number of iterations per element of `parameters`
              if(sum(lapply(
                  object@parameters[setdiff(names(object@parameters), c("RefFreq", "designMatrix"))],
                  nrow) != N) > 0)
                errors <- c(errors, "Different numbers of iterations")
              # Check dimensions for basic gene-specific parameters
              if(sum(lapply(object@parameters[c("mu", "delta")], ncol) != q) > 0)
                errors <- c(errors, 
                            "Parameters' dimensions are not compatible (genes)")
              # Check dimensions for basic cell-specific parameters
              if(sum(lapply(object@parameters[c("s", "nu")], ncol) != n) > 0)
                errors <- c(errors, 
                            "Parameters' dimensions are not compatible (cells)")
              # Check dimensions for optional gene-specific parameters
              if("epsilon" %in% names(object@parameters))
              {
                if((nrow(object@parameters$epsilon) != N) |
                   (ncol(object@parameters$epsilon) != q))
                  errors <- c(errors, 
                              "Invalid dimensions for `epsilon`")                  
              }
              # Check dimensions for optional cell-specific parameters
              if("phi" %in% names(object@parameters))
              {
                if((nrow(object@parameters$phi) != N) |
                   (ncol(object@parameters$phi) != n) |
                   (sum(!is.finite(object@parameters$phi)) > 0))
                  errors <- c(errors, 
                              "Invalid dimensions for `phi`")                  
              }
              # Check labels for gene-specific params
              if(sum(!all.equal(colnames(object@parameters$mu),
                                colnames(object@parameters$delta))) > 0)
              {
                errors <- c(errors,"Gene-specific parameter labels don't match")
              }
              
              if (length(errors) == 0) 
                TRUE else errors  
            }
            
      },
      prototype = prototype(new("Versioned",
                                versions = c("BASiCS_Chain" = 
                                              utils::packageVersion("BASiCS"))))
      )


#' @name BASiCS_Summary
#' @aliases BASiCS_Summary-class
#' 
#' @title The BASiCS_Summary class
#' 
#' @description Container of a summary of a \code{\linkS4class{BASiCS_Chain}} 
#' object. In each element of the \code{parameters} slot, first column contains 
#' posterior medians; second and third columns respectively contain the lower 
#' and upper limits of an high posterior density interval (for a given probability).
#' 
#' @slot parameters List of parameters in which each entry contains a matrix: 
#' first column contains posterior medians, second column contains the 
#' lower limits of an high posterior density interval and third column 
#' contains the upper limits of high posterior density intervals. 
#' \describe{
#' \item{mu}{Posterior medians (1st column), lower (2nd column) and upper 
#' (3rd column) limits of gene-specific mean expression parameters \eqn{\mu_i}.}
#' \item{delta}{Posterior medians (1st column), lower (2nd column) and upper 
#' (3rd column) limits of gene-specific biological over-dispersion parameters 
#' \eqn{\delta_i}, biological genes only }
#' \item{phi}{Posterior medians (1st column), lower (2nd column) and upper 
#' (3rd column) limits of cell-specific mRNA content normalisation 
#' parameters \eqn{\phi_j}}
#' \item{s}{Posterior medians (1st column), lower (2nd column) and upper 
#' (3rd column) limits of cell-specific technical normalisation 
#' parameters \eqn{s[j]}}
#' \item{nu}{Posterior medians (1st column), lower (2nd column) and upper 
#' (3rd column) limits of cell-specific random effects \eqn{\nu_j}}
#' \item{theta}{Posterior median (1st column), lower (2nd column) and upper 
#' (3rd column) limits of technical over-dispersion parameter(s) \eqn{\theta} 
#' (each row represents one batch)}
#' \item{\code{beta}}{Posterior median (first column), lower (second column) 
#' and upper (third column) limits of regression coefficients \eqn{\beta}}
#' \item{\code{sigma2}}{Posterior median (first column), lower (second column) 
#' and upper (third column) limits of residual variance \eqn{\sigma^2}}
#' \item{\code{epsilon}}{Posterior median (first column), lower (second column) 
#' and upper (third column) limits of gene-specific residual over-dispersion 
#' parameter \eqn{\epsilon} }
#' }
#' 
#' @examples
#' 
#' # A BASiCS_Summary object created by the Summary method.
#' Data <- makeExampleBASiCS_Data()
#' Chain <- BASiCS_MCMC(Data, N = 100, Thin = 2, Burn = 2, Regression = FALSE)
#' ChainSummary <- Summary(Chain)
#' 
#' 
#' @export
setClass("BASiCS_Summary", 
         representation = representation(
           parameters = "list"
         ), 
         contains="Versioned",
         validity = function(object) 
        {
           errors <- character()
           
           ValidNames <- c("mu", "delta", "phi", "s", "nu", "theta",
                           "beta", "sigma2", "epsilon")
           ReqNames <- c("mu", "delta", "s", "nu", "theta")
           
           # Check whether all elements of `parameters` are valid
           if(sum(!(names(object@parameters) %in% ValidNames) > 0) > 0)
             errors <- c(errors, "Invalid elemens in `parameters` slot")
           
           # Check whether all minimum `parameters` are present
           if(sum(names(object@parameters) %in% ReqNames) != 5)
             errors <- c(errors, "One or more parameters are missing")     
           
           if (length(errors) == 0) 
             TRUE else errors 
      },
      prototype = prototype(new("Versioned",
                                versions = c("BASiCS_Summary" = 
                                               utils::packageVersion("BASiCS"))))
)
